unit module Terminal::ANSI:ver<0.0.6>;

constant CSI = "\e[";

sub nums(*@n) {
 # trim trailing nils?
 @n.map({$_ // ''}).join(';')
}

sub csi($str) {
  print-or-buffer CSI ~ $str;
}

sub esc($str) {
  print-or-buffer "\e" ~ $str;
}

sub print-or-buffer($str) {
  if ($*terminal-ansi-buffer) {
    $*terminal-ansi-out ~= $str;
  } else {
    print $str;
  }
}


#| Clear the screen.
sub clear-screen is export {
  csi '2J';
}

#| Move the cursor to line, column.
sub move-to($l,$c = 1) is export {
  csi nums($l,$c) ~ "H";
}

#| Move to home (0,0).
sub home is export {
  csi 'H';
}

my $scroll-bottom;
my $scroll-top;

#| Set scroll region to top, bottom.
sub set-scroll-region($top,$bottom) is export {
  $scroll-bottom = $bottom;
  $scroll-top = $top;
  csi nums($top,$bottom) ~ "r";
}

#| Reset the scroll region.
sub reset-scroll-region is export {
  $scroll-bottom = Nil;
  $scroll-top = Nil;
  csi 'r';
}

#| Scroll down 1.
sub scroll-down is export {
  move-to $scroll-top, 1 if $scroll-top;
  esc "M";
}

#| Scroll up 1.
sub scroll-up is export {
  move-to $scroll-bottom, 1 if $scroll-bottom;
  esc "D";
}

#| Hide the cursor.
sub hide-cursor is export {
  csi '?25l';
}

#| Show the cursor.
sub show-cursor is export {
  csi "?25h";
}

#| Save screen state (smcup).
sub save-screen is export {
  csi '?1049h'
}

#| Restore screen (rmcup).
sub restore-screen is export {
  csi '?1049l'
}

#| Turn off cursor (civis).
sub cursor-off is export {
  csi '?25l'
}

#| Turn on cursor (cnorm).
sub cursor-on is export {
  csi '?25h';
}

#| Move cursor down.
sub cursor-down($amt = 1) is export {
  csi $amt ~ 'B';
}

#| Move cursor up.
sub cursor-up($amt = 1) is export {
  csi $amt ~ 'A';
}

#| Move cursor right.
sub cursor-right($amt = 1) is export {
  csi $amt ~ 'C';
}

#| Move cursor left.
sub cursor-left($amt = 1) is export {
  csi $amt ~ 'D';
}

#| Atomic move + print.
sub print-at($r,$c,$str) is export {
  csi nums($r,$c) ~ "H" ~ $str;
}

#| Set fg color to $n.
sub set-fg-color($n) is export {
  csi "38;5;" ~ $n ~ "m";
}

#| Set bg color to $n.
sub set-bg-color($n) is export {
  csi "48;5;" ~ $n ~ "m";
}

#| Save the cursor position.
sub save-cursor is export {
  csi ~ 's';
}

#| restore the cursor position.
sub restore-cursor is export {
  csi ~ 'u';
}

#| Move to start of line.
sub start-of-line is export {
  print-or-buffer "\r"
}

#| Erase to end of line.
sub erase-to-end-of-line is export {
  csi '2J';
}

#| Reverse video
sub reverse-video is export {
  csi '7m';
}

#| Normal video
sub normal-video is export {
  csi '0m';
}

#| Bold (increased intensity)
sub bold is export {
  csi '1m';
}

#| Faint (lower intensity)
sub faint is export {
  csi '2m';
}

#| Get strings instead of printing them.
sub tget(&callable) is export {
  my $*terminal-ansi-buffer = True;
  my $*terminal-ansi-out = "";
  callable;
  $*terminal-ansi-out;
}

#| Combine a series of outputs into one.
sub atomically(&callable) is export {
  die "cannot nest atomically" with CALLERS::{'$*terminal-ansi-out'};
  my $*terminal-ansi-out = '';
  my $*terminal-ansi-buffer = True;
  callable;
  print $*terminal-ansi-out;
}

#| Return a description of the given input sequence
sub parse-input($str) is export {
  my %inputs =
   "\e[A" => "Up",
   "\e[B" => "Down",
   "\e[C" => "Right",
   "\e[D" => "Left",
   "\e[H" => "Home",
   "\e[F" => "End",
   "\e[G" => "Keypad 5",
   "\e[1P" => "F1",
   "\e[1Q" => "F2",
   "\e[1R" => "F3",
   "\e[1S" => "F4",

   "\e[10~" => "F0",
   "\e[11~" => "F1",
   "\e[12~" => "F2",
   "\e[13~" => "F3",
   "\e[14~" => "F4",
   "\e[15~" => "F5",

   "\e[17~" => "F6",
   "\e[18~" => "F7",
   "\e[19~" => "F8",
   "\e[20~" => "F9",
   "\e[21~" => "F10",

   "\e[23~" => "F11",
   "\e[24~" => "F12",
   "\e[25~" => "F13",
   "\e[26~" => "F14",

   "\e[28~" => "F15",
   "\e[29~" => "F16",

   "\e[31~" => "F17",
   "\e[32~" => "F18",
   "\e[33~" => "F19",
   "\e[34~" => "F20";

  .return with %inputs{ $str };
  if $str.chars == 2 && $str.starts-with("\e") {
    return "ALT + " ~ $str.substr(1,1);
  }

  if $str.starts-with("\e[1;") && $str.chars > 5 {
    my $modifier = $str.substr(4,1) - 1;
    my @mods;
    @mods.push: "META" if $modifier +& 8;
    @mods.push: "ALT" if $modifier +& 2;
    @mods.push: "CTRL" if $modifier +& 4;
    @mods.push: "SHIFT" if $modifier +& 1;
    with %inputs{ "\e[" ~ $str.substr(5) } -> $key {
      return @mods.join(" + ") ~ " + $key";
    }
  }
  Nil
}
